#include "sipl.h"
#include <stdlib.h> 
#include <stdio.h>
#include <math.h>

struct img * imgCreate(int m, int n) {
	struct img* img1 = (struct img*) malloc(sizeof(struct img));
	img1 -> row = m;
	img1 -> col = n;
	img1->r = (unsigned char *) calloc(sizeof(unsigned char), m * n);
	img1->g = (unsigned char *) calloc(sizeof(unsigned char), m * n);
	img1->b = (unsigned char *) calloc(sizeof(unsigned char), m * n);
	return img1;
}

struct matrix * matCreate(int m, int n) {
	struct matrix* mat = (struct matrix*) malloc(sizeof(struct matrix));
	mat->row = m;
	mat->col = n;
	mat->data = (double *)calloc(sizeof(double), m * n);
	return mat;
}

struct matrix * initMatrix(double * mat_data_ptr, int m, int n) {
	struct matrix* mat = (struct matrix*) malloc(sizeof(struct matrix));
	mat->row = m;
	mat->col = n;
	mat->data = mat_data_ptr;
	return mat;
}

double matAccess(struct matrix * mat, int i, int j) {
	if (mat->row <= i) i = mat->row - 1;
	if (mat->col <= j) j = mat->col - 1;
	return mat->data[i * mat->col + j];
}

int matAssign(struct matrix * mat, int i, int j, double val) {
	if (mat->row <= i) i = mat->row - 1;
	if (mat->col <= j) j = mat->col - 1;
	mat->data[i * mat->col + j] = val;
	return 0;
}

struct pixel * pixAccess(struct img * img1, int i, int j) {
	struct pixel* pix = (struct pixel*) malloc(sizeof(struct pixel));
	if (img1->row <= i) i = img1->row - 1;
	if (img1->col <= j) j = img1->col - 1;
	pix->r = img1->r[i * img1->col + j];
	pix->g = img1->g[i * img1->col + j];
	pix->b = img1->b[i * img1->col + j];
	return pix;
}

int pixAssign(struct img * img1, int i, int j, struct pixel* pix) {
	if (img1->row <= i) i = img1->row - 1;
	if (img1->col <= j) j = img1->col - 1;
	img1->r[i * img1->col + j] = (unsigned char) (pix->r);
	img1->g[i * img1->col + j] = (unsigned char) (pix->g);
	img1->b[i * img1->col + j] = (unsigned char) (pix->b);
	return 0;
}

int imgAccess(struct img * img1, int i, int j, const char* s) {
	if (img1->row <= i) i = img1->row - 1;
	if (img1->col <= j) j = img1->col - 1;
    if(strcmp(s, "r") == 0)
	    return (int)img1->r[i * img1->col + j];
    if(strcmp(s, "g") == 0)
	    return (int)img1->g[i * img1->col + j];
    if(strcmp(s, "b") == 0)
	    return (int)img1->b[i * img1->col + j];
	return 0;
}

int imgAssign(struct img * img1, int i, int j, const char* s, int val) {
	if (img1->row <= i) i = img1->row - 1;
	if (img1->col <= j) j = img1->col - 1;
	if (val > 255) {
		val = 255;
	} else if (val < 0) {
		val = 0;
	}
    if(strcmp(s, "r") == 0)
	    img1->r[i * img1->col + j] = (unsigned char) (val);
    if(strcmp(s, "g") == 0)
	    img1->g[i * img1->col + j] = (unsigned char) (val);
    if(strcmp(s, "b") == 0)
	    img1->b[i * img1->col + j] = (unsigned char) (val);
	return 0;
}

int imgWidth(struct img* img1){

    return img1->col;
}

int matWidth(struct matrix *mat){
    return mat->col;
}

int imgHeight(struct img* img1){

    return img1->row;
}

int matHeight(struct matrix *mat){
    return mat->row;
}

double int2float(int n) {
	return (double) n;
}

int float2int (double n) {
	return (int) n;
}

double power (double x, double y) {
	return pow(x,y);
}
int printMatrix(struct matrix * mat) {
	int i, j;
	//printf("%d\n", mat->row);
	//printf("%d\n", mat->col);
	for (i = 0; i < mat->row; i++) {
		for (j = 0; j < mat->col; j++) {
			printf("%f",mat->data[i * mat->col + j]);
		}
		printf("\n");
	}
	return 0;
}

int printPixel (struct pixel * pix) {
	printf("%d,%d,%d\n", pix->r, pix->g, pix->b);
	return 0;
}

int changeGrey(struct img* image){
	int i;
	unsigned char avg;
	for(i = 0; i < image->col * image->row; i++){
		avg = (image->r[i] + image->g[i] + image->b[i]) / 3;
		image->r[i] = avg;
		image->g[i] = avg;
		image->b[i] = avg;
		
	}
	return 0;
/*	img1->r -= img1->col * img1->row;
	img1->g -= img1->col * img1->row;
	img1->b -= img1->col * img1->row;
	*/
}


//rotate clockwise
int rotateImage(struct img* image){
	int y, x, i;
	const int width = image->col;
	const int height = image->row;
	unsigned char temp_matrix[height * width * 3];
	
	//use a new matrix to store
	for(i = 0; i < width * height; i++){
		temp_matrix[3 * i] = image->r[i];
		temp_matrix[3 * i + 1] = image->g[i];
		temp_matrix[3 * i + 2] = image->b[i];
	}
	
	
	
	for(y = 0; y < width; y++){
		for(x = 0; x < height; x++){
			image->r[y * height + x] = temp_matrix[3 * ((height - x - 1) * width + (width - y - 1))];	
			image->g[y * height + x] = temp_matrix[3 * ((height - x - 1) * width + (width - y - 1)) + 1] ;	
			image->b[y * height + x] = temp_matrix[3 * ((height - x - 1) * width + (width - y - 1)) + 2];	
		}
	}
	
	image->row = width;
	image->col = height; 
	return 0;
}



int flipImage(struct img* image, int n){
	if(n != -1 && n!= 1) return 1;
	int x, y;
	unsigned char tmpr, tmpg, tmpb;
	
	if(n == 1){
		for(y = 0; y < image -> row; y++){
			for(x = 0; x < image -> col / 2; x++){
				tmpr = image->r[y * image -> col + x];
				image->r[y * image -> col + x] = image->r[y * image -> col + image -> col - x - 1];
				image->r[y * image -> col + image -> col - x - 1] = tmpr;
				
				tmpg = image->g[y * image -> col + x];
				image->g[y * image -> col + x] = image->g[y * image -> col + image -> col - x - 1];
				image->g[y * image -> col + image -> col - x - 1] = tmpg;
				
				tmpb = image->b[y * image -> col + x];
				image->b[y * image -> col + x] = image->b[y * image -> col + image -> col - x - 1];
				image->b[y * image -> col + image -> col - x - 1] = tmpb;
			}		
		}			
	}	
	
	else {
		for(x = 0; x < image -> col; x++){
			for(y = 0; y < image -> row / 2; y++){
				tmpr = image->r[y * image -> col + x];
				image->r[y * image -> col + x] = image->r[(image -> row - y - 1) * image -> col + x];
				image->r[(image -> row - y - 1) * image -> col + x] = tmpr;
				
				tmpg = image->g[y * image -> col + x];
				image->g[y * image -> col + x] = image->g[(image -> row - y - 1) * image -> col + x];
				image->g[(image -> row - y - 1) * image -> col + x] = tmpg;
				
				tmpb = image->b[y * image -> col + x];
				image->b[y * image -> col + x] = image->b[(image -> row - y - 1) * image -> col + x];
				image->b[(image -> row - y - 1) * image -> col + x] = tmpb;
			}		
		}
	}
	
	return 0;
}


//create kernel matrix
struct matrix * kernel(const char* str){
	int i;
	struct matrix* mat = (struct matrix*) malloc(sizeof(struct matrix));

	if(strcmp("edge", str) == 0){
		mat->data = (double *)malloc(sizeof(double) * 9);
		for(i = 0; i < 9; i++){
			if( i == 4) mat->data[i] = 8;
			else mat->data[i] = -1;
		}
		
	mat->col = mat->row = 3;
	}
	
	else if(strcmp("sharpen", str) == 0){
		mat->data = (double *)malloc(sizeof(double) * 9);
		for(i = 0; i < 9; i++){
			if(i == 4) mat->data[i] = 5;
			else if(i % 2 == 1) mat->data[i] = -1;
			else mat->data[i] = 0;
		}
		
		mat->col = mat->row = 3;
	}
	
	else if(strcmp("gaussian", str) == 0){
		mat->data = (double *)malloc(sizeof(double) * 9);
		
		for(i = 0; i < 9; i++){
			if(i == 4) mat->data[i] = 0.25;
			else if(i % 2 == 1) mat->data[i] = 0.125;
			else mat->data[i] = 1/16;
		}
		
		mat->col = mat->row = 3;
	}

	return mat;	
	
}


struct img * convImg(struct img *image, struct matrix *mat){
	
	int x, y;
	int i, j;
	const int width = image->col;
	const int height = image->row;
	
	//rows and cols of matrix must be odd
	if(mat->col % 2 == 0 || mat->row % 2 == 0) return image;
	
	
	double* temp_mat; //[height * width * 3] = {0};
	temp_mat = (double *)malloc(sizeof(double) * 3 * height * width);
	memset(temp_mat, 0, 3 * height * width * sizeof(double));
	
	
	
	//save the result in tmp matrix
	for(y = 0; y < image->row; y++)	{
		
		//if it is at the edge of picture
		if( y - mat->row / 2 < 0 || y + mat->row / 2 >= image->row)   continue;
		
		for(x = 0; x < image->col; x++){
			
			if( x - mat->col / 2 < 0 || x + mat->col / 2 >= image->col) continue;
			
			for( i = -mat->row / 2; i <= mat->row / 2; i++){
				for(j = - mat->col / 2; j <= mat->col / 2; j++){
					temp_mat[ 3 * (y * width + x)] 		+= (double)image->r[(y + i) * width + x + j] * mat->data[(i + mat->row / 2) * mat->col + j + mat->col / 2];
					temp_mat[ 3 * (y * width + x) + 1] 	+= (double)image->g[(y + i) * width + x + j] * mat->data[(i + mat->row / 2) * mat->col + j + mat->col / 2];
					temp_mat[ 3 * (y * width + x) + 2] 	+= (double)image->b[(y + i) * width + x + j] * mat->data[(i + mat->row / 2) * mat->col + j + mat->col / 2];
					
				} 
			}
		
		}
	}
	
	//give the value to image
	for(y = 0; y < image->row; y++)	{
		
		//if it is at the edge of picture
		if( y - mat->row / 2 < 0 || y + mat->row / 2 >= image->row)   continue;
		for(x = 0; x < image->col; x++){
			
			if( x - mat->col / 2 < 0 || x + mat->col / 2 >= image->col) continue;
			for(i = 0; i < 3; i++){
				if(temp_mat[3 * (y * width + x) + i] > 255 ) temp_mat[3 * (y * width + x) + i] = 255;
				else if (temp_mat[3 * (y * width + x) + i] < 0 ) temp_mat[3 * (y * width + x) + i] = 0;
			}
			
			image->r[y * width + x] = (unsigned char) temp_mat[3 * (y * width + x)]; 
			image->g[y * width + x] = (unsigned char) temp_mat[3 * (y * width + x) + 1]; 
			image->b[y * width + x] = (unsigned char) temp_mat[3 * (y * width + x) + 2]; 
		}
	}
	
	free(temp_mat);
	
	return image;
}

struct img* imgAdd(struct img *img1,struct img *img2 ){
	struct img* img3;
	img3 = (struct img *)malloc(sizeof(struct img));
	img3->r = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->g = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->b = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->col = img1->col;
	img3->row = img1->row;
	
	int i = 0;
	//unsigned char tmp; 
	for(i = 0; i < img1->col * img1->row; i++){
		if((int)img1->r[i] + (int)img2 -> r[i] <= 255)
			img3->r[i] = img1->r[i] + img2 -> r[i];
		else 
			img3->r[i] = 255;
		if((int)img1->g[i] + (int)img2 -> g[i] <= 255)
			img3->g[i] = img1->g[i] + img2 -> g[i];
		else 
			img3->g[i] = 255;
		if((int)img1->b[i] + (int)img2 -> b[i] <= 255)
			img3->b[i] = img1->b[i] + img2 -> b[i];
		else 
			img3->b[i] = 255;	

		
	}
	
	
	return img3;
}


struct img* imgCopy(struct img *img1){
	struct img* img3;
	img3 = (struct img *)malloc(sizeof(struct img));
	img3->r = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->g = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->b = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->col = img1->col;
	img3->row = img1->row;
	
	int i = 0;
	//unsigned char tmp; 
	for(i = 0; i < img1->col * img1->row; i++){
		
		img3->r[i] = img1->r[i];
	
		img3->g[i] = img1->g[i];
		
		img3->b[i] = img1->b[i];
		
	}
	
	return img3;
}

struct img* elementAdd(struct img *img1, int num2 ){
	/*struct img* img1;
	img3 = ( img *)malloc(sizeof(img3));
	img1->r = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img1->g = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img1->b = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img1->col = img1->col;
	img1->row = img1->row;
	*/
	unsigned char num;
	num = (unsigned char) num2;
	// printf("num: %d \n");
	int i = 0;
	//unsigned char tmp; 
	for(i = 0; i < img1->col * img1->row; i++){
		
		//printf("int r %d  sum:%d", (int)(img1->r[i]), (int)(img1->r[i]) + num2 );
		//printf("int g %d  sum:%d", (int)(img1->g[i]), (int)(img1->g[i]) + num2 );
		//printf("int b %d  sum:%d", (int)(img1->b[i]), (int)(img1->b[i]) + num2 );
		
		if((img1->r[i]) + num2 <= 255)
			img1->r[i] = img1->r[i] + num;
		else{
			//printf("r %d \n", i); 
			img1->r[i] = 255;
		}
			
			
		if((img1->g[i]) + num2 <= 255)
			img1->g[i] = img1->g[i] + num;
		else{
			//printf("g %d \n", i); 
			img1->g[i] = 255;
		}
			
			
		if((img1->b[i]) + num2 <= 255)
			img1->b[i] = img1->b[i] + num;
		else{
			//printf("b %d \n", i); 
			img1->b[i] = 255;
		}
			
	
		
	}
	
	//printf("address %d  r : %d", img1, img1->r);
	return img1;
}


struct img* elementSub(struct img *img1, int num2 ){
	/*struct img* img3;
	img3 = (struct img *)malloc(sizeof(img3));
	img3->r = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->g = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->b = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->col = img1->col;
	img3->row = img1->row;
	*/
	int i = 0;
	unsigned char num = (unsigned char)num2;
	//unsigned char tmp; struct
	for(i = 0; i < img1->col * img1->row; i++){
		if(img1->r[i] - num2 > 0)
			img1->r[i] = img1->r[i] - num;
		else
			img1->r[i] = 0;
			
		if(img1->g[i] - num2 > 0)
			img1->g[i] = img1->g[i] - num;
		else
			img1->g[i] = 0;
			
		if(img1->b[i] - num2 > 0)
			img1->b[i] = img1->b[i] - num;
		else
			img1->b[i] = 0;
		
	}
	
	return img1;
}


struct img* elementMult(struct img *img1, double num ){
	/*
	struct img* img3;
	img3 = (struct img *)malloc(sizeof(img3));
	img3->r = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->g = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->b = (unsigned char *) malloc(sizeof(unsigned char) * img1->col * img1->row);
	img3->col = img1->col;
	img3->row = img1->row;
	*/
	int i = 0;
	//unsigned char tmp; 
	for(i = 0; i < img1->col * img1->row; i++){
		
		img1->r[i] = (unsigned char) img1->r[i] * num;
		img1->g[i] = (unsigned char)img1->g[i] * num;
		img1->b[i] = (unsigned char)img1->b[i] * num;
		
	}
	return img1;
}

struct pixel * pixAddPix(struct pixel * pix1, struct pixel * pix2) {
	if (pix1->r + pix2->r < 255) {
		pix1->r = pix1->r + pix2->r;
	} else {
		pix1->r = 255;
	}
	if (pix1->g + pix2->g < 255) {
		pix1->g = pix1->g + pix2->g;
	} else {
		pix1->g = 255;
	}
	if (pix1->b + pix2->b < 255) {
		pix1->b = pix1->b + pix2->b;
	} else {
		pix1->b = 255;
	}
	return pix1;
}

struct pixel * pixAddInt(struct pixel * pix1, int val) {
	if (pix1->r + val > 255) {
		pix1->r = 255;
	} else if (pix1->r + val < 0) {
		pix1->r = 0;
	} else {
		pix1->r = (unsigned char) pix1->r + val;
	}
	if (pix1->g + val > 255) {
		pix1->g = 255;
	} else if (pix1->g + val < 0) {
		pix1->g = 0;
	} else {
		pix1->g = (unsigned char) pix1->g + val;
	}
	if (pix1->b + val > 255) {
		pix1->b = 255;
	} else if (pix1->b + val < 0) {
		pix1->b = 0;
	} else {
		pix1->b = (unsigned char) pix1->b + val;
	}
	return pix1;
}

struct pixel * pixSubInt(struct pixel * pix1, int val) {
	if (pix1->r - val > 255) {
		pix1->r = 255;
	} else if (pix1->r - val < 0) {
		pix1->r = 0;
	} else {
		pix1->r = (unsigned char) pix1->r - val;
	}
	if (pix1->g - val > 255) {
		pix1->g = 255;
	} else if (pix1->g - val < 0) {
		pix1->g = 0;
	} else {
		pix1->g = (unsigned char) pix1->g - val;
	}
	if (pix1->b - val > 255) {
		pix1->b = 255;
	} else if (pix1->b - val < 0) {
		pix1->b = 0;
	} else {
		pix1->b = (unsigned char) pix1->b - val;
	}
	return pix1;
}


struct pixel * pixMultFloat(struct pixel * pix1, double val) {
	if (pix1->r * val > 255) {
		pix1->r = 255;
	} else if (pix1->r * val < 0) {
		pix1->r = 0;
	} else {
		pix1->r = (unsigned char) pix1->r * val;
	}
	if (pix1->g * val > 255) {
		pix1->g = 255;
	} else if (pix1->g * val < 0) {
		pix1->g = 0;
	} else {
		pix1->g = (unsigned char) pix1->g * val;
	}
	if (pix1->b * val > 255) {
		pix1->b = 255;
	} else if (pix1->b * val < 0) {
		pix1->b = 0;
	} else {
		pix1->b = (unsigned char) pix1->b * val;
	}
	return pix1;
}